/*-
 * Copyright (c) 2006 Verdens Gang AS
 * Copyright (c) 2006-2008 Linpro AS
 * All rights reserved.
 *
 * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: cache_pipe.c 3717 2009-02-10 14:20:29Z tfheen $
 *
 * XXX: charge bytes to srcaddr
 */

#include "config.h"

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <poll.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include "shmlog.h"
#include "cache.h"
#include "vend.h"

static int
rdf(int fd0, int fd1)
{
	int i, j;
	char buf[BUFSIZ], *p;

	i = read(fd0, buf, sizeof buf);
	if (i <= 0)
		return (1);
	for (p = buf; i > 0; i -= j, p += j) {
		j = write(fd1, p, i);
		if (j <= 0)
			return (1);
		if (i != j)
			(void)usleep(100000);		/* XXX hack */
	}
	return (0);
}


union proxy_addr {
	struct {        /* for TCP/UDP over IPv4, len = 12 */
		uint32_t src_addr;
		uint32_t dst_addr;
		uint16_t src_port;
		uint16_t dst_port;
	} ip4;
	struct {        /* for TCP/UDP over IPv6, len = 36 */
		uint8_t  src_addr[16];
		uint8_t  dst_addr[16];
		uint16_t src_port;
		uint16_t dst_port;
	} ip6;
};


struct proxy_hdr_v2 {
	uint8_t sig[12];  /* hex 0D 0A 0D 0A 00 0D 0A 51 55 49 54 0A */
	uint8_t ver_cmd;  /* protocol version and command */
	uint8_t fam;      /* protocol family and address */
	uint16_t len;     /* number of following bytes part of the header */
	union proxy_addr addr;
};

/*
 * to use WRW_ / writev, we cannot have the extent to send on the
 * stack, so we construct our proxy_addr on workspace
 */
static inline size_t
proxyhdr(struct worker *w, struct ws *ws, struct sockaddr *server,
    struct sockaddr *client)
{
	size_t l;
	struct proxy_hdr_v2 *hdr;

	/* over-allocing for the ipv4 case */
	hdr = WS_Malloc(ws, sizeof(*hdr));
	if (hdr == NULL)
		return 0;

	hdr->sig[0]  = '\r';
	hdr->sig[1]  = '\n';
	hdr->sig[2]  = '\r';
	hdr->sig[3]  = '\n';
	hdr->sig[4]  = '\0';
	hdr->sig[5]  = '\r';
	hdr->sig[6]  = '\n';
	hdr->sig[7]  = 'Q';
	hdr->sig[8]  = 'U';
	hdr->sig[9]  = 'I';
	hdr->sig[10] = 'T';
	hdr->sig[11] = '\n';

	hdr->ver_cmd = 0x21; /* PROXY2 PROXY */

	/* we just hardcode SOCK_STREAM - 0x1 in fam */
	switch (client->sa_family) {
	case AF_INET:
		hdr->fam = 0x11;
		hdr->len = 12;
		hdr->addr.ip4.src_addr =
		    ((struct sockaddr_in *)client)->sin_addr.s_addr;
		hdr->addr.ip4.src_port =
		    ((struct sockaddr_in *)client)->sin_port;
		hdr->addr.ip4.dst_addr =
		    ((struct sockaddr_in *)server)->sin_addr.s_addr;
		hdr->addr.ip4.dst_port =
		    ((struct sockaddr_in *)server)->sin_port;
		break;
	case AF_INET6:
		hdr->fam = 0x21;
		hdr->len = 36;
		memcpy(hdr->addr.ip6.src_addr,
		    &((struct sockaddr_in6 *)client)->sin6_addr, 16);
		hdr->addr.ip6.src_port =
		    ((struct sockaddr_in6 *)client)->sin6_port;
		memcpy(hdr->addr.ip6.dst_addr,
		    &((struct sockaddr_in6 *)server)->sin6_addr, 16);
		hdr->addr.ip6.dst_port =
		    ((struct sockaddr_in6 *)server)->sin6_port;
		break;
	default:
		INCOMPL();
	}

	l = sizeof(*hdr) - sizeof(union proxy_addr) + hdr->len;

	vbe16enc(&hdr->len, hdr->len);

	return WRW_Write(w, hdr, l);
}

void
PipeSession(struct sess *sp, const int proxy)
{
	struct vbe_conn *vc;
	struct worker *w;
	struct bereq *bereq;
	struct pollfd fds[2];
	int i;

	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
	CHECK_OBJ_NOTNULL(sp->wrk, WORKER_MAGIC);
	w = sp->wrk;
	bereq = sp->bereq;
	sp->bereq = NULL;

	VBE_GetFd(sp);
	if (sp->vbe == NULL)
		return;
	vc = sp->vbe;
	if (TCP_blocking(vc->fd)) {
		vca_close_session(sp, "pipe");
		VBE_ClosedFd(sp);
		return;
	}

	WRW_Reserve(w, &vc->fd);

	if (proxy) {
		if (sp->mysockaddr->sa_family == AF_UNSPEC)
			assert(!getsockname(sp->fd, sp->mysockaddr, &sp->mysockaddrlen)
			|| errno == ECONNRESET);

		i = proxyhdr(w, sp->ws, sp->mysockaddr, sp->sockaddr);
		if (i == 0) {
			vca_close_session(sp, "pipe - proxy");
			VBE_ClosedFd(sp);
			return;
		}
		sp->acct_req.hdrbytes += i;
	}

	sp->acct_req.hdrbytes += http_Write(w, bereq->http, 0);

	if (sp->htc->pipeline.b != NULL)
		sp->acct_req.bodybytes +=
		    WRW_Write(w, sp->htc->pipeline.b, Tlen(sp->htc->pipeline));

	i = WRW_FlushRelease(w);
	VBE_free_bereq(&bereq);

	if (i) {
		vca_close_session(sp, "pipe");
		VBE_ClosedFd(sp);
		return;
	}

	sp->t_resp = TIM_real();

	memset(fds, 0, sizeof fds);
	fds[0].fd = vc->fd;
	fds[0].events = POLLIN | POLLERR;
	fds[1].fd = sp->fd;
	fds[1].events = POLLIN | POLLERR;

	while (fds[0].fd > -1 || fds[1].fd > -1) {
		fds[0].revents = 0;
		fds[1].revents = 0;
		i = poll(fds, 2, params->pipe_timeout * 1000);
		if (i < 1)
			break;
		if (fds[0].revents && rdf(vc->fd, sp->fd)) {
			(void)shutdown(vc->fd, SHUT_RD);
			(void)shutdown(sp->fd, SHUT_WR);
			fds[0].events = 0;
			fds[0].fd = -1;
		}
		if (fds[1].revents && rdf(sp->fd, vc->fd)) {
			(void)shutdown(sp->fd, SHUT_RD);
			(void)shutdown(vc->fd, SHUT_WR);
			fds[1].events = 0;
			fds[1].fd = -1;
		}
	}
	if (fds[0].fd >= 0) {
		(void)shutdown(vc->fd, SHUT_RD);
		(void)shutdown(sp->fd, SHUT_WR);
	}
	if (fds[1].fd >= 0) {
		(void)shutdown(sp->fd, SHUT_RD);
		(void)shutdown(vc->fd, SHUT_WR);
	}
	vca_close_session(sp, "pipe");
	VBE_ClosedFd(sp);
}
